/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <ctype.h>
#include <values.h>

#include <qdir.h> 
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qtextstream.h>

#include <kapp.h>
#include <kconfig.h>
#include <kfiledialog.h>
#include <klocale.h>

#include "dialog.h"
#include "dialogmargin.h"
#include "exportdialog.h"


CExportDialog::CExportDialog( const QString &title, QWidget *parent, 
			      char *name, bool modal )
  : CPageDialog( CPageDialog::TABS,  CPageDialog::HELP |
		 CPageDialog::OK | CPageDialog::CANCEL,
		 CPageDialog::OK, title, parent, name, modal )
{

  
  mFrame[ page_destination ] = addPage( -1, i18n("Destination"), 
					i18n("Destination") );
  mFrame[ page_option ] = addPage( -1, i18n("Options"), i18n("Options") );

  setupDestinationPage();
  setupOptionPage();

  adjustSize();
  showPage( page_destination );
 
  mConfig = 0;
  readConfiguration();

  QString path = mDestination.fileInput->text();
  int index = path.findRev( '/' );
  if( index != -1 ) { mWorkDir = path.left( index+1 ); }

}


CExportDialog::~CExportDialog( void )
{
  writeConfiguration();
  delete mConfig; mConfig = 0;
}


void CExportDialog::readConfiguration( void )
{
  if( mConfig != 0 ) { return; }

  QString configPath( kapp->localconfigdir() );
  QString configFile( "/hexexport" );

  mConfig = new KSimpleConfig( configPath + configFile );
  if( mConfig == 0 ) { return; }

  mConfig->setGroup( "Destination" );

  int val = mConfig->readNumEntry( "Format", 0 );
  mDestination.formatCombo->setCurrentItem( val );
  formatChanged( val < 0 || val > option_html ? 0 : val );

  mConfig->setGroup( "Option" );
  val = mConfig->readNumEntry( "HtmlLine", 80 );
  mOption.lineSpin->setValue( val );
  QString text = mConfig->readEntry( "HtmlPrefix", "table" );
  mOption.prefixInput->setText( text );
  val = mConfig->readNumEntry( "HtmlHeader", 1 );
  mOption.topCombo->setCurrentItem( val < 0 || val >= 4 ? 0 : val );
  val = mConfig->readNumEntry( "HtmlFooter", 3 );
  mOption.bottomCombo->setCurrentItem( val < 0 || val >= 4 ? 0 : val );

  bool state = mConfig->readBoolEntry( "HtmlSymlink", true );
  mOption.symlinkCheck->setChecked( state );  
  state = mConfig->readBoolEntry( "HtmlNavigatorBar", true );
  mOption.navigatorCheck->setChecked( state );
  state = mConfig->readBoolEntry( "HtmlBlackWhite", false );
  mOption.bwCheck->setChecked( state );
  
}


void CExportDialog::writeConfiguration( void )
{
  if( mConfig == 0 )
  {
    return;
  }

  mConfig->setGroup( "Destination" );
  mConfig->writeEntry( "Format", mDestination.formatCombo->currentItem() );

  mConfig->setGroup( "Option" );
  mConfig->writeEntry( "HtmlLine", mOption.lineSpin->value() );
  mConfig->writeEntry( "HtmlPrefix", mOption.prefixInput->text() );
  mConfig->writeEntry( "HtmlHeader", mOption.topCombo->currentItem() );
  mConfig->writeEntry( "HtmlFooter", mOption.bottomCombo->currentItem() );
  
  mConfig->writeEntry( "HtmlSymlink", mOption.symlinkCheck->isChecked());
  mConfig->writeEntry( "HtmlNavigatorBar",mOption.navigatorCheck->isChecked());
  mConfig->writeEntry( "HtmlBlackWhite", mOption.bwCheck->isChecked() );

  mConfig->sync();
}


void CExportDialog::setupDestinationPage( void )
{
  QString text;
  QFrame *page = mFrame[ page_destination ];

  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );  
  if( topLayout == 0 ) { return; }
  

  const char *items[] = 
  { 
    i18n("Plain text"),
    i18n("Html tables"),
    0 
  };
  mDestination.formatCombo = new QComboBox( false, page );
  mDestination.formatCombo->insertStrList( items );
  mDestination.formatCombo->setMinimumWidth( fontMetrics().maxWidth()*10 );
  mDestination.formatCombo->setFixedHeight( 
    mDestination.formatCombo->sizeHint().height() );
  connect( mDestination.formatCombo, SIGNAL(activated(int)), 
	   SLOT(formatChanged(int)) );
  
  text = i18n("&Format"); 
  QLabel *label = new QLabel( mDestination.formatCombo, text, page );
  label->setMinimumSize( label->sizeHint() );
  topLayout->addWidget( label, 0, AlignLeft );
  topLayout->addWidget( mDestination.formatCombo );



  QHBoxLayout *hbox = new QHBoxLayout();
  if( hbox == 0 ) { return; }
  topLayout->addLayout( hbox );

  text = i18n("&Destination"); 
  label = new QLabel( text, page );
  label->setMinimumSize( label->sizeHint() );
  hbox->addWidget( label );
 
  text = i18n("(Package directory)"); 
  mDestination.fileExtraLabel = new QLabel( text, page );
  mDestination.fileExtraLabel->setMinimumSize( 
    mDestination.fileExtraLabel->sizeHint() );
  hbox->addWidget( mDestination.fileExtraLabel );
  hbox->addStretch(10);



  hbox = new QHBoxLayout();
  if( hbox == 0 ) { return; }
  topLayout->addLayout( hbox );

  mDestination.fileInput = new QLineEdit( page );
  hbox->addWidget( mDestination.fileInput, 10 );

  text = i18n("&Browse");
  QPushButton *browseButton = new QPushButton( text, page, "browse" );
  browseButton->setFixedSize( browseButton->sizeHint() );
  hbox->addWidget( browseButton );
  connect( browseButton, SIGNAL(clicked()), SLOT(browserClicked()) );
  mDestination.fileInput->setFixedHeight( browseButton->sizeHint().height() );
  mDestination.fileInput->setMinimumWidth( fontMetrics().maxWidth()*15 );

  label->setBuddy(mDestination.fileInput);



  mDestination.rangeBox = new QButtonGroup( i18n("Export range"), page );
  if( mDestination.rangeBox == 0 ) { return; }
  topLayout->addWidget( mDestination.rangeBox );

  QButtonGroup *group = mDestination.rangeBox; // convenience

  QVBoxLayout *vbox = new QVBoxLayout( group, INNER_SEP, INNER_SEP );  
  vbox->addSpacing( fontMetrics().lineSpacing() );
  
  mDestination.radio1 = new QRadioButton( i18n("&Everything"), group );
  mDestination.radio1->setFixedSize( mDestination.radio1->sizeHint() );
  mDestination.rangeBox->insert( mDestination.radio1, 0 );
  vbox->addWidget( mDestination.radio1, 0, AlignLeft );

  mDestination.radio2 = new QRadioButton( i18n("&Selection"), group );
  mDestination.radio2->setFixedSize( mDestination.radio2->sizeHint() );
  mDestination.rangeBox->insert( mDestination.radio2, 1 );
  vbox->addWidget( mDestination.radio2, 0, AlignLeft );

  mDestination.radio3 = new QRadioButton( i18n("&Range"), group );
  mDestination.radio3->setFixedSize( mDestination.radio3->sizeHint() );
  mDestination.rangeBox->insert( mDestination.radio3, 2 );
  vbox->addWidget( mDestination.radio3, 0, AlignLeft );

  QGridLayout *gbox = new QGridLayout( 2, 2, INNER_SEP );
  vbox->addLayout( gbox );
  gbox->setColStretch( 1, 10 );

  mDestination.fromInput = new QLineEdit( group );
  mDestination.fromInput->setFixedHeight( 
    mDestination.fromInput->sizeHint().height() );
  text = i18n("&From offset"); 
  mDestination.fromLabel = new QLabel( mDestination.fromInput, text, group );  
  mDestination.fromLabel->setMinimumSize( mDestination.fromLabel->sizeHint() );
  gbox->addWidget( mDestination.fromLabel, 0, 0 );
  gbox->addWidget( mDestination.fromInput, 0, 1 );

  mDestination.toInput = new QLineEdit( group );
  mDestination.toInput->setFixedHeight( 
    mDestination.toInput->sizeHint().height() );
  text = i18n("&To offset"); 
  mDestination.toLabel = new QLabel( mDestination.toInput, text, group );
  mDestination.toLabel->setMinimumSize( mDestination.toLabel->sizeHint() );
  gbox->addWidget( mDestination.toLabel, 1, 0 );
  gbox->addWidget( mDestination.toInput, 1, 1 );

  connect( group, SIGNAL(clicked(int)), SLOT(rangeChanged(int)) );
  group->setButton(0);
  rangeChanged(0);


  vbox->activate();

  topLayout->addStretch(10);
  topLayout->activate();

}


void CExportDialog::setupOptionPage( void )
{
  QFrame *page = mFrame[ page_option ];
  QVBoxLayout *topLayout = new QVBoxLayout( page, OUTER_SEP, INNER_SEP );  
  if( topLayout == 0 ) { return; }

  mOptionStack = new QWidgetStack( page, "stack" );
  if( mOptionStack == 0 ) { return; }
  topLayout->addWidget( mOptionStack );

  makeTextOption();
  makeHtmlOption();
  mOptionStack->raiseWidget( (int)option_text );

  topLayout->activate();

  QSize size;
  for( int i=0; i<page_max; i++ )
  {
    QWidget *w = mOptionStack->widget(i);
    if( w != 0 )
    {
      size = size.expandedTo( w->minimumSize() );
    }
  }
  page->setMinimumSize( size + QSize(OUTER_SEP*2, OUTER_SEP*2) );
}


void CExportDialog::makeTextOption( void )
{
  QFrame *page = new QFrame( mFrame[ page_option ] );
  if( page == 0 ) { return; }
  mOptionStack->addWidget( page, option_text );
}


void CExportDialog::makeHtmlOption( void )
{
  QFrame *page = new QFrame( mFrame[ page_option ] );
  if( page == 0 ) { return; }
  mOptionStack->addWidget( page, option_html );

  QString text;
  QVBoxLayout *topLayout = new QVBoxLayout( page, 0, INNER_SEP );  
  if( topLayout == 0 ) { return; }


  text = i18n("Html options. One table per page");
  QLabel *label = new QLabel( text, page );
  label->setMinimumSize( label->sizeHint() );
  topLayout->addWidget( label, 0, AlignLeft );

  QFrame *hline = new QFrame( page );
  hline->setFrameStyle( QFrame::Sunken | QFrame::HLine );
  hline->setFixedHeight( hline->sizeHint().height() );
  topLayout->addWidget( hline );


  QFrame *frame = new QFrame( page );
  if( frame == 0 ) { return; }  
  topLayout->addWidget( frame );

  QGridLayout *gbox = new QGridLayout( frame, 4, 2, 0, INNER_SEP );
  if( gbox == 0 ) { return; }
  gbox->setColStretch( 1, 10 );  

  mOption.lineSpin = new QSpinBox( frame );
  mOption.lineSpin->setMinimumWidth( fontMetrics().maxWidth()*10 );
  mOption.lineSpin->setFixedHeight( mOption.lineSpin->sizeHint().height() );
  mOption.lineSpin->setRange( 5, MAXINT );
  gbox->addWidget( mOption.lineSpin, 0, 1 );
  
  text = i18n("&Lines per table");
  label = new QLabel( mOption.lineSpin, text, frame );
  label->setMinimumSize( label->sizeHint() );
  gbox->addWidget( label, 0, 0 );

  mOption.prefixInput = new QLineEdit( frame, "prefix" );
  mOption.prefixInput->setMinimumWidth( fontMetrics().maxWidth()*10 );
  mOption.prefixInput->setFixedHeight( 
    mOption.prefixInput->sizeHint().height() );
  gbox->addWidget( mOption.prefixInput, 1, 1 );
  
  text = i18n("Filename prefix (in package)");
  label = new QLabel( mOption.prefixInput, text, frame );
  label->setMinimumSize( label->sizeHint() );
  gbox->addWidget( label, 1, 0 );


  mOption.topCombo = new QComboBox( false, frame );
  const char *items[] = 
  { 
    i18n("None"),
    i18n("Filename with path"),
    i18n("Filename"),
    i18n("Page number"),
    0 
  };
  mOption.topCombo->insertStrList( items );
  mOption.topCombo->setFixedHeight( 
    mOption.topCombo->sizeHint().height() );
  mOption.topCombo->setMinimumWidth(
    mOption.topCombo->sizeHint().width() + 10 );
  gbox->addWidget( mOption.topCombo, 2, 1 );

  text = i18n("Header &above text");
  label = new QLabel( mOption.topCombo, text, frame );
  label->setMinimumSize( label->sizeHint() );
  gbox->addWidget( label, 2, 0 );

  mOption.bottomCombo = new QComboBox( false, frame );
  mOption.bottomCombo->insertStrList( items );
  mOption.bottomCombo->setFixedHeight( 
    mOption.bottomCombo->sizeHint().height() );
  mOption.bottomCombo->setMinimumWidth(
    mOption.bottomCombo->sizeHint().width() + 10 );
  gbox->addWidget( mOption.bottomCombo, 3, 1 );

  text = i18n("&Footer below text");
  label = new QLabel( mOption.bottomCombo, text, frame );
  label->setMinimumSize( label->sizeHint() );
  gbox->addWidget( label, 3, 0 );

  gbox->activate();

  text = i18n("Link \"index.html\" to &table of contents file");
  mOption.symlinkCheck = new QCheckBox( text, page );
  mOption.symlinkCheck->setMinimumSize( mOption.symlinkCheck->sizeHint() );
  topLayout->addWidget( mOption.symlinkCheck );

  text = i18n("&Include navigator bar");
  mOption.navigatorCheck = new QCheckBox( text, page );
  mOption.navigatorCheck->setMinimumSize( mOption.navigatorCheck->sizeHint() );
  topLayout->addWidget( mOption.navigatorCheck, 0, AlignLeft );

  text = i18n("&Use black and white only");
  mOption.bwCheck = new QCheckBox( text, page );
  mOption.bwCheck->setMinimumSize( mOption.bwCheck->sizeHint() );
  topLayout->addWidget( mOption.bwCheck, 0, AlignLeft );

  topLayout->addStretch(10);
  topLayout->activate();
}



void CExportDialog::formatChanged( int index )
{
  mDestination.formatCombo->setCurrentItem(index);  
  if( index == 0 )
  {
    mDestination.fileExtraLabel->setEnabled( false );
    mOptionStack->raiseWidget( (int)option_text );
  }
  else
  {
    mDestination.fileExtraLabel->setEnabled( true );
    mOptionStack->raiseWidget( (int)option_html );
  }

}


void CExportDialog::rangeChanged( int id )
{
  bool state = id == 2 ? true : false;
  mDestination.toLabel->setEnabled( state );
  mDestination.fromLabel->setEnabled( state );
  mDestination.toInput->setEnabled( state );
  mDestination.fromInput->setEnabled( state );
}


void CExportDialog::browserClicked( void )
{
  QString url;
  if( mDestination.formatCombo->currentItem() == option_html )
  {
    url = KFileDialog::getDirectory( mWorkDir, topLevelWidget() );
  }
  else
  {
    url = KFileDialog::getSaveFileName( mWorkDir, "*", topLevelWidget());
  }

  if( url.isEmpty() == true )
  {
    return;
  }
  else
  {
    int index = url.findRev( '/' );
    if( index != -1 ) { mWorkDir = url.left( index+1 ); }
  }

  mDestination.fileInput->setText( url );
}


void CExportDialog::slotHelp( void )
{
  kapp->invokeHTMLHelp( QString("khexedit/khexedit.html"), QString() );
}



void CExportDialog::slotOk( void )
{
  QString path( mDestination.fileInput->text() );

  int format = mDestination.formatCombo->currentItem();
  if( format == option_text )
  {
    if( verifyFileDestnation( this, i18n("Export Document"), path ) == false )
    {
      return;
    }

    SExportText e;
    uint mode;
    if( collectRange( mode, e.range.start, e.range.stop ) == false )
    {
      showEntryFailure( this, QString("") );
      return;
    }
    e.range.mode = (SExportRange::EMode)mode; // FIXME
    e.destFile   = path;

    hide();
    emit exportText(e);
  }
  else if( format == option_html )
  {
    SExportHtml e;
    uint mode;
    if( collectRange( mode, e.range.start, e.range.stop ) == false )
    {
      showEntryFailure( this, QString("") );
      return;
    }
 
    e.range.mode = (SExportRange::EMode)mode;

    QString str = mOption.prefixInput->text();
    str.stripWhiteSpace();
    if( str.isEmpty() == true )
    {
      mOption.prefixInput->setText( "table" );
    }

    const QString prefix = mOption.prefixInput->text();
    for( uint i=0; i<prefix.length(); i++ )
    {
      int c = prefix.data()[i];
      if( isspace(c) || ispunct(c) )
      {
	QString msg = i18n(""
          "The filename prefix can not contain empty letters\n"
	  "or punctations");

	QMessageBox::warning(
	  this,
	  i18n("Export Document"),
	  msg,
	  i18n("&OK"),
	  "",
	  "",
	  0,0 );
	return;
      }
    }

    if( verifyPackage( path ) == false )
    {
      return;
    }

    e.package       = path;
    e.prefix        = prefix;
    e.linePerPage   = mOption.lineSpin->value();
    e.topCaption    = mOption.topCombo->currentItem();
    e.bottomCaption = mOption.bottomCombo->currentItem();
    e.symLink       = mOption.symlinkCheck->isChecked();
    e.navigator     = mOption.navigatorCheck->isChecked();
    e.blackWhite    = mOption.bwCheck->isChecked();

    hide();
    emit exportHtml(e);
  }

}




bool CExportDialog::collectRange( uint &mode, uint &start, uint &stop )
{
  if( mDestination.radio1->isChecked() == true )
  {
    mode = SExportRange::All;
    return( true );
  }
  else if( mDestination.radio2->isChecked() == true )
  {
    mode = SExportRange::Selection;
    return( true );
  }
  else if( mDestination.radio3->isChecked() == true )
  {
    mode = SExportRange::Range;
    bool ok1 = stringToOffset( mDestination.fromInput->text(), start );
    bool ok2 = stringToOffset( mDestination.toInput->text(), stop );
    if( ok1 == false || ok2 == false || start >= stop )
    {
      return( false );
    }
    return( true );
  }

  return( false );
}


//
// This one will attempt to create a directory if 'path'
// specifies a nonexistent name.
//
bool CExportDialog::verifyPackage( const QString &path )
{
  const QString title = i18n("Export Document");

  if( path.isEmpty() == true )
  {
    QString msg = i18n("You must specify a destination.");
    QMessageBox::warning(
      this,
      title,
      msg,
      i18n("&OK"),
      "",
      "",
      0,0 );
    return( false );
  }

  QFileInfo info( path );
  if( info.exists() == false )
  {
    QDir directory;
    if( directory.mkdir( path ) == false )
    {
      QString msg;
      msg += i18n("Unable to create new a new directory");
      msg += "\n";
      msg += path;

      QMessageBox::warning(
        this,
        title,
        msg,
        i18n("&OK"),
        "",
        "",
        0,0 );
      return( false );
    }
  }
  else
  {
    if( info.isDir() == false ) 
    {
      QString msg = i18n("You have specified an existing file");
      QMessageBox::warning(
        this,
	title,
	msg,
	i18n("&OK"),
	"",
	"",
	0,0 );
      return( false );
    }
    else
    {
      if( info.isWritable() == false ) 
      {
	QString msg = i18n( ""
          "You do not have write permission to this directory.");

	QMessageBox::warning(
          this,
	  title,
	  msg,
	  i18n("&OK"),
	  "",
	  "",
	  0,0 );
	return( false );
      }

      const char *prefix = mOption.prefixInput->text();
      QString f1, f2;
      f1.sprintf("%s00000000.html", prefix == 0 ? "" : prefix );
      f2.sprintf("%s99999999.html", prefix == 0 ? "" : prefix );

      QString msg = i18n( ""
        "You have specified an existing directory.\n"
	"If you continue, any existing file in the range" );
      msg += "\n";
      msg += f1;
      msg += " ";
      msg += i18n("to");
      msg += " ";
      msg += f2;
      msg += "\n";
      msg += i18n("can be lost.");
      msg += "\n\n";
      msg += i18n("Continue?");

      int reply = QMessageBox::warning( 
        this,
        title,
	msg,
        i18n( "&Yes" ),
        i18n( "&No" ),
        "",
        0, 1 );
      if( reply != 0 )
      {
        return( false );
      }
    }
  }

  return( true );
}
